"""
Общие функции клиента и сервера
"""
import json
from common.variables import MAX_PACKAGE_LENGTH, ENCODING
from common.decos import Log
from common.errors import NotDictError


@Log()
def send_message(socket_obj, message):
    """
    Функция, осуществляющая кодирование и отправку сообщений между
    клиентами
    :param socket_obj: объект сокета для обмена сообщениями
    :param message: словарь с атрибутами сообщения
    """
    if not isinstance(message, dict):
        raise NotDictError
    msg = json.dumps(message)
    socket_obj.send(msg.encode(ENCODING))


@Log()
def get_message(socket_obj):
    """
    Функция принимает и декодирует сообщение
    :param socket_obj: объект сокета для обмена сообщениями
    :return: словарь с атрибутами сообщения
    """
    response = socket_obj.recv(MAX_PACKAGE_LENGTH)
    # проверяеи пришедшие данные
    if isinstance(response, bytes):
        response = json.loads(response.decode(ENCODING))
        # Проверяем результат декодирования
        if isinstance(response, dict):
            return response
        raise NotDictError
    raise ValueError



"""
База данных для клиентской стороны мессенджера.
БД содержит следующие таблицы:
a) список всех пользователей;
b) список контактов;
c) история сообщений.
Использует SQLite базу данных, реализован с помощью
SQLAlchemy ORM с использованием декларативного подхода.
"""
import os
from sqlalchemy import create_engine, Column, Integer, String, DateTime, or_
from sqlalchemy.orm import sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from datetime import datetime


class ClientStorage:
    """
    Класс-оболочка для клиентской базы данных.
    """
    Base = declarative_base()

    class AllUsers(Base):
        """
        Таблица "Все пользователеи"
        """
        __tablename__ = 'all_users'

        id = Column(Integer, primary_key=True)
        user_name = Column(String, unique=True)

        def __init__(self, name):
            self.user_name = name

    class ContactList(Base):
        """
        Таблица "Список контактов"
        """
        __tablename__ = 'contact_list'

        id = Column(Integer, primary_key=True)
        user_name = Column(String, unique=True)

        def __init__(self, name):
            self.user_name = name

    class MessageHistory(Base):
        """
        Таблица "История сообщений"
        """
        __tablename__ = 'message_history'

        id = Column(Integer, primary_key=True)
        sender = Column(String)
        recipient = Column(String)
        msg_text = Column(String)
        msg_time = Column(DateTime)

        def __init__(self, sender, recipient, msg_text):
            self.sender = sender
            self.recipient = recipient
            self.msg_text = msg_text
            self.msg_time = datetime.now()

    def __init__(self, name):
        path = os.path.dirname(os.path.realpath(__file__))
        filename = f'client_{name}.db3'
        self.engine = create_engine(f'sqlite:///{os.path.join(path, filename)}',
                                    echo=False,
                                    pool_recycle=7200,
                                    connect_args={'check_same_thread': False})

        self.Base.metadata.create_all(self.engine)
        Session = sessionmaker(self.engine)
        self.session = Session()

        self.session.query(self.ContactList).delete()
        self.session.commit()

    def save_message(self, sender, recipient, msg_text):
        """
        Функция сохраниения сообщений.
        """
        self.session.add(self.MessageHistory(sender, recipient, msg_text))
        self.session.commit()

    def get_message_history(self, name=None):
        """
        Получает историю сообщений с пользователем или всю историю сообщений
        """
        history = self.session.query(self.MessageHistory)
        if name:
            history = history.filter(or_(self.MessageHistory.sender == name,
                                         self.MessageHistory.recipient == name))

        result = [(row.sender, row.recipient, row.msg_time, row.msg_text) for row in history]
        return result

    def add_all_users(self, user_list):
        """
        Добавляет в базу список пользователей, полученный с сервера.
        """
        self.session.query(self.AllUsers).delete()
        for user in user_list:
            user_row = self.AllUsers(user)
            self.session.add(user_row)
        self.session.commit()

    def get_users(self):
        """
        Возвращает список известных пользователей.
        """
        return [user[0] for user in self.session.query(self.AllUsers.user_name).all()]

    def check_user(self, user):
        """
        Проверяющяя наличие пользователя в известных.
        """
        if self.session.query(self.AllUsers).filter_by(user_name=user).count():
            return True
        else:
            return False

    def add_contacts(self, contact_list):
        """
        Обновляет список контактов.
        """
        self.session.query(self.ContactList).delete()
        for user in contact_list:
            self.session.add(self.ContactList(user))
        self.session.commit()

    def add_contact(self, name):
        """
        Добавляет контакт в список контактов
        """
        if not self.session.query(self.ContactList).filter_by(user_name=name).count():
            self.session.add(self.ContactList(name))
            self.session.commit()

    def del_contact(self, name):
        """
        Удаляет контакт из списка контактов
        """
        self.session.query(self.ContactList).filter_by(user_name=name).delete()
        self.session.commit()

    def get_contacts(self):
        """
        Возвращает список контактов пользователя
        """
        return [contact[0] for contact in self.session.query(self.ContactList.user_name).all()]

    def check_contact(self, contact):
        """
        Проверяет наличие пользователя в списке контактов
        """
        if self.session.query(self.ContactList).filter_by(user_name=contact).count():
            return True
        else:
            return False


if __name__ == '__main__':
    test_db = ClientStorage('test1')
    for i in ['test3', 'test4', 'test5']:
        test_db.add_contact(i)

    test_db.add_contact('test4')

    test_db.add_contacts(['test1', 'test2', 'test3', 'test4', 'test5'])

    test_db.save_message('test1', 'test2', 'тестовое сообщение')
    test_db.save_message('test2', 'test1', 'другое тестовое сообщение')
    print(test_db.get_contacts())

    print(test_db.check_contact('test1'))
    print(test_db.check_contact('test10'))

    for msg in test_db.get_message_history('test2'):
        print(f'От {msg[0]} для {msg[1]} в {msg[2]}\n'
              f'{msg[3]}')

    for msg in test_db.get_message_history():
        print(f'От {msg[0]} для {msg[1]} в {msg[2]}\n'
              f'{msg[3]}')

    test_db.del_contact('test4')
    print(test_db.get_contacts())



from PyQt5.QtWidgets import QDialog, QPushButton, QLineEdit, QLabel
from PyQt5.QtWidgets import QApplication, qApp


class UserNameDialog(QDialog):
    """
    Класс диалогового окна для ввода имени пользователя и пароля
    """
    def __init__(self):
        super().__init__()

        self.ok_pressed = False
        self.setWindowTitle('Подключение')
        self.setFixedSize(210, 140)

        self.name_label = QLabel('Введите имя пользователя:', self)
        self.name_label.move(10, 10)
        self.name_label.setFixedSize(150, 10)

        self.client_name = QLineEdit(self)
        self.client_name.setFixedSize(190, 20)
        self.client_name.move(10, 30)

        self.pass_label = QLabel('Введите пароль:', self)
        self.pass_label.move(10, 60)
        self.pass_label.setFixedSize(150, 10)

        self.client_pass = QLineEdit(self)
        self.client_pass.setFixedSize(190, 20)
        self.client_pass.move(10, 80)
        self.client_pass.setEchoMode(QLineEdit.Password)

        self.btn_ok = QPushButton('ОК', self)
        self.btn_ok.move(30, 110)
        self.btn_ok.clicked.connect(self.click)

        self.btn_cancel = QPushButton('Выход', self)
        self.btn_cancel.move(110, 110)
        self.btn_cancel.clicked.connect(qApp.exit)

        self.show()

    def click(self):
        """Обработчик нажатия кнопки ОК"""
        if self.client_name.text() and self.client_pass.text():
            self.ok_pressed = True
            qApp.exit()


if __name__ == '__main__':
    app = QApplication([])
    dial = UserNameDialog()
    app.exec_()



"""
Конфигурационный файл для логгирования клиента
"""
import logging
import os

CLIENT_LOGGER = logging.getLogger('client')
LOG_FILE_NAME = os.path.join(os.path.dirname(__file__), "client_log.log")

FORMATTER = logging.Formatter("%(asctime)s - %(levelname)-8s - %(filename)-10s - %(message)s")

FILE_HANDLER = logging.FileHandler(LOG_FILE_NAME, encoding='utf-8')
FILE_HANDLER.setFormatter(FORMATTER)

STREAM_HANDLER = logging.StreamHandler()
STREAM_HANDLER.setFormatter(FORMATTER)
STREAM_HANDLER.setLevel(logging.ERROR)

CLIENT_LOGGER.addHandler(FILE_HANDLER)
CLIENT_LOGGER.addHandler(STREAM_HANDLER)
CLIENT_LOGGER.setLevel(logging.DEBUG)

if __name__ == '__main__':
    CLIENT_LOGGER.debug('Test. Debug info')
    CLIENT_LOGGER.info('Test. Information')
    CLIENT_LOGGER.warning('Test. Warning')
    CLIENT_LOGGER.error(f'Test. Error')
    CLIENT_LOGGER.critical(f'Test. Critical error')



"""
Unit-тесты для модуля utils.py
"""

import json
import os
import sys
import unittest
from socket import socket, AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR

sys.path.append(os.path.join(os.getcwd(), '..'))
from common.variables import *
from common.utils import get_message, send_message
from common.errors import NotDictError


class TestUtils(unittest.TestCase):
    test_message = {
        ACTION: PRESENCE,
        TIME: 1,
        USER: {
            'account_name': 'User',
            'password': ''
        }
    }
    test_correct_response = {
        RESPONSE: 200,
        TIME: 1,
        ALERT: 'Соединение прошло успешно'
    }
    test_error_response = {
        RESPONSE: 400,
        TIME: 1,
        ERROR: 'Ошибка соединения'
    }

    # инициализируем тестовые сокеты для клиента и для сервера
    server_socket = None
    client_socket = None

    def setUp(self) -> None:
        # Создаем тестовый сокет для сервера
        self.server_socket = socket(AF_INET, SOCK_STREAM)
        self.server_socket.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self.server_socket.bind((DEFAULT_LISTEN_ADDRESSES, DEFAULT_PORT))
        self.server_socket.listen(MAX_USERS)
        # Создаем тестовый сокет для клиента
        self.client_socket = socket(AF_INET, SOCK_STREAM)
        self.client_socket.connect((DEFAULT_IP, DEFAULT_PORT))

    def tearDown(self) -> None:
        # Закрываем созданные сокеты
        self.client_socket.close()
        self.server_socket.close()

    def test_send_wrong_message_from_client(self):
        """
        Проверяем исключение, если на входе не словарь
        """
        self.assertRaises(NotDictError, send_message, self.client_socket, 'not dict')

    def test_send_message_client_server(self):
        """
        Проверяем отправку корректного сообщения
        """
        client, client_address = self.server_socket.accept()
        # Отправляем сообщение
        send_message(self.client_socket, self.test_message)
        # Получаем и раскодируем сообщение
        test_response = client.recv(MAX_PACKAGE_LENGTH)
        test_response = json.loads(test_response.decode(ENCODING))
        client.close()
        # Проверяем соответствие изначального сообщения и прошедшего отправку
        self.assertEqual(self.test_message, test_response)

    def test_get_message_200(self):
        """
        Корректрая расшифровка коректного словаря
        """
        # Отправляем клиенту тестовый ответ о корректной отправке данных
        client, client_address = self.server_socket.accept()
        message = json.dumps(self.test_correct_response)
        client.send(message.encode(ENCODING))
        client.close()
        # получаем ответ
        response = get_message(self.client_socket)
        # сравниваем отправленный и полученный ответ
        self.assertEqual(self.test_correct_response, response)

    def test_get_message_400(self):
        """
        Корректрая расшифровка ошибочного словаря
        """
        # Отправляем клиенту тестовый ответ об ошибке
        client, client_address = self.server_socket.accept()
        message = json.dumps(self.test_error_response)
        client.send(message.encode(ENCODING))
        client.close()
        # получаем ответ
        response = get_message(self.client_socket)
        # сравниваем отправленный и полученный ответ
        self.assertEqual(self.test_error_response, response)

    def test_get_message_not_dict(self):
        """
        Проверяем возникновение ошибки, если пришедший объект не словарь
        """
        client, client_address = self.server_socket.accept()
        # Отправляем клиенту строку, вместо словаря
        message = json.dumps('not dict')
        client.send(message.encode(ENCODING))
        client.close()

        self.assertRaises(NotDictError, get_message, self.client_socket)

    def test_get_message_dict(self):
        """
        Проверяет является ли возвращаемый объект словарем
        """
        client, client_address = self.server_socket.accept()
        message = json.dumps(self.test_correct_response)
        client.send(message.encode(ENCODING))
        client.close()

        self.assertIsInstance(get_message(self.client_socket), dict)


if __name__ == '__main__':
    unittest.main()



"""
Настройки по умолчанию и описание
JIM-протокола для передачи сообщений
"""

DEFAULT_PORT = 7777
DEFAULT_IP = '127.0.0.1'
DEFAULT_LISTEN_ADDRESSES = ''
TIMEOUT = 0.5

MAX_PACKAGE_LENGTH = 4096
MAX_USERS = 10

ENCODING = 'utf-8'

# JIM-протокол
ACTION = 'action'
TIME = 'time'
TO = 'to'
FROM = 'from'
TEXT = 'message'
USER = 'user'
LOGIN = 'user_login'
RESPONSE = 'response'
ALERT = 'alert'
ERROR = 'error'
DATA = 'bin'
ACCOUNT_NAME = 'account_name'
PUBLIC_KEY = 'pubkey'

# Действия (actions)
PRESENCE = 'presence'
MSG = 'msg'
EXIT = 'quit'
GET_CONTACTS = 'get_contacts'
GET_USERS = 'get_users'
ADD_CONTACT = 'add_contact'
DEL_CONTACT = 'del_contact'
PUBLIC_KEY_REQUEST = 'pubkey_need'

# Ответы сервера
RESPONSE_200 = {
    RESPONSE: 200,
    ALERT: None
}

RESPONSE_202 = {
    RESPONSE: 202,
    ALERT: None
}

RESPONSE_400 = {
    RESPONSE: 400,
    ERROR: None
}

RESPONSE_511 = {
    RESPONSE: 511,
    DATA: None
}



"""
Ошибки, которые могут возникнуть при работе сервера и клиента
"""


class NotDictError(Exception):
    def __str__(self):
        return 'Аргумент функции должен быть словарем'


class ServerError(Exception):
    def __init__(self, text):
        self.text = text

    def __str__(self):
        return self.text



"""
Декораторы
"""

import logging
import inspect
import log.client_log_config
import log.server_log_config
from sys import argv
from functools import wraps


class Log:
    """
    Класс-декоратор для логгирования вызова функций
    """
    def __init__(self):
        # определяем логгер
        if argv[0].find('run_server.py') != -1:
            self.func_logger = logging.getLogger('server')
        else:
            self.func_logger = logging.getLogger('client')

    def __call__(self, function):
        # возвращаем имя и док-стринг декорируемой функции
        @wraps(function)
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            self.func_logger.debug(f'Вызвана функция {function.__name__} с параметрами: {args} {kwargs}.')
            self.func_logger.debug(f'Функция {function.__name__} вызвана из функции {inspect.stack()[1][3]}')
            return result
        return wrapper



"""
Модуль для запуска клиентской части мессенджера.
параметры командной строки:
addr          — ip-адрес сервера;
port          — tcp-порт на сервере, по умолчанию 7777;
-n --name     — имя пользователя;
-p --password — пароль.
"""
import argparse
import logging
import os
from sys import argv, exit

from Cryptodome.PublicKey import RSA
from PyQt5.QtWidgets import QApplication

from client.client_db import ClientStorage
from client.main_window import ClientMainWindow
from client.start_dialog import UserNameDialog
from client.transport import MessengerClient
from common.decos import Log
from common.errors import ServerError
from common.variables import *

client_log = logging.getLogger('client')


@Log()
def get_client_settings():
    """
    Получает параметры клиента из командной сторки
    или назначает параметры по умолчанию.
    Возвращает порт и ip-адрес сервера, имя пользователя и пароль
    """
    args = argparse.ArgumentParser()
    args.add_argument('address', default=DEFAULT_IP, nargs='?')
    args.add_argument('port', type=int, default=DEFAULT_PORT, nargs='?')
    args.add_argument('-n', '--name', default=None, nargs='?')
    args.add_argument('-p', '--password', default='', nargs='?')
    namespace = args.parse_args(argv[1:])
    connection_ip = namespace.address
    connection_port = namespace.port
    user_name = namespace.name
    user_password = namespace.password

    if not (1024 < connection_port < 65535):
        client_log.critical(f'Неверное значение порта {connection_port}.\n'
                            f'Порт должен находиться в диапазоне от 1024 до 65535.')
        exit(1)

    client_log.debug(f'Получены параметры подключения {connection_ip}:{connection_port}')

    return connection_ip, connection_port, user_name, user_password


@Log()
def main():
    """
    Основная функция для запуска клиентской части мессенджера.
    """
    conn_ip, conn_port, user_name, user_password = get_client_settings()

    client_app = QApplication(argv)

    if not user_name or not user_password:
        start_dialog = UserNameDialog()
        client_app.exec_()
        if start_dialog.ok_pressed:
            user_name = start_dialog.client_name.text()
            user_password = start_dialog.client_pass.text()
            del start_dialog
        else:
            exit(0)

    dir_path = os.path.dirname(os.path.realpath(__file__))
    key_file = os.path.join(dir_path, f'{user_name}.key')
    if not os.path.exists(key_file):
        keys = RSA.generate(2048, os.urandom)
        with open(key_file, 'wb') as key:
            key.write(keys.export_key())
    else:
        with open(key_file, 'rb') as key:
            keys = RSA.import_key(key.read())

    client_db = ClientStorage(user_name)

    # Создаём подключение
    try:
        connection = MessengerClient(user_name, user_password, conn_ip, conn_port, client_db, keys)
    except ServerError as error:
        print(error.text)
        exit(1)
    connection.setDaemon(True)
    connection.start()

    client_log.info(f'Запущен клиент для пользователя {user_name}.'
                    f'Адрес подключения: {conn_ip}, порт:{conn_port}')

    # GUI
    main_window = ClientMainWindow(client_db, connection, keys)
    main_window.make_connection(connection)
    main_window.setWindowTitle(f'Добро пожаловать, {user_name}')
    client_app.exec_()

    connection.connection_shutdown()
    connection.join()


if __name__ == '__main__':
    main()



import binascii
import hashlib
import hmac
import json
import logging
import sys
import threading
from socket import socket, AF_INET, SOCK_STREAM
from time import time, sleep

from PyQt5.QtCore import pyqtSignal, QObject

sys.path.append('../')
from common.utils import send_message, get_message
from common.variables import *
from common.decos import Log
from common.errors import ServerError

client_log = logging.getLogger('client')
sock_lock = threading.Lock()


class MessengerClient(threading.Thread, QObject):
    """
    Класс-поток для сообщения с сервером.
    """
    new_message = pyqtSignal(dict)
    connection_lost = pyqtSignal()

    def __init__(self, user_name, password, connection_ip, connection_port, database, keys):
        threading.Thread.__init__(self)
        QObject.__init__(self)

        self.user_name = user_name
        self.password = password
        self.keys = keys
        self.db = database
        self.socket = None
        self.create_connection(connection_ip, connection_port)

        # Обновляем таблицы известных пользователей и контактов
        try:
            self.get_user_list()
            self.get_contact_list()
        except OSError as err:
            if err.errno:
                client_log.critical(f'Потеряно соединение с сервером.')
                raise ServerError('Потеряно соединение с сервером.')
            client_log.error('Timeout соединения при обновлении списков пользователей.')
        except json.JSONDecodeError:
            client_log.critical(f'Потеряно соединение с сервером.')
            raise ServerError('Потеряно соединение с сервером.')
        self.running = True

    @Log()
    def create_connection(self, ip, port):
        """
        Функция пытается установить соединение с сервером
        из полученных ip-адреса и порта и отправляет на сервер запрос
        на авторизацию.
        """
        try:
            self.socket = socket(AF_INET, SOCK_STREAM)
            self.socket.settimeout(5)
            self.socket.connect((ip, port))

        except (ConnectionRefusedError, ConnectionError):
            client_log.critical(f'Не удалось установить соединение с сервером {ip}:{port}')
            raise ServerError('Не удалось установить соединение с сервером')

        else:
            client_log.info(f'Соединение с сервером {ip}:{port}')
            client_log.info('Запуск процедуры авторизации.')

            password_hash = hashlib.pbkdf2_hmac('sha512',
                                                self.password.encode(ENCODING),
                                                self.user_name.encode(ENCODING),
                                                10000)
            passwd_hash_string = binascii.hexlify(password_hash)
            pubkey = self.keys.publickey().export_key().decode('ascii')

            message = self.create_presence_message(pubkey)

            try:
                with sock_lock:
                    send_message(self.socket, message)
                    answer = get_message(self.socket)
                    if RESPONSE in answer:
                        if answer[RESPONSE] == 400:
                            client_log.error(f'Ошибка авторизации. {answer[ERROR]}')
                            raise ServerError(answer[ERROR])
                        elif answer[RESPONSE] == 511:
                            data = answer[DATA]
                            digest = hmac.new(passwd_hash_string, data.encode(ENCODING), 'MD5').digest()
                            my_answer = RESPONSE_511
                            my_answer[DATA] = binascii.b2a_base64(digest).decode('ascii')
                            send_message(self.socket, my_answer)
                            self.read_response(get_message(self.socket))
                            client_log.info('Авторизация прошла успешно.')
            except (OSError, json.JSONDecodeError):
                client_log.critical('Сбой соединения в процессе авторизации.')
                raise ServerError('Сбой соединения в процессе авторизации.')

    @Log()
    def create_presence_message(self, pubkey):
        """
        Функция формирует presence-сообщение для сервера.
        :param pubkey: публичный ключ пользователя.
        """
        message = {
            ACTION: PRESENCE,
            TIME: time(),
            USER: {
                ACCOUNT_NAME: self.user_name,
                PUBLIC_KEY: pubkey
            }
        }
        client_log.debug(f'Создано приветственное сообщение серверу от {self.user_name}')
        return message

    @Log()
    def get_contact_list(self):
        """
        Функция отправляет запрос серверу на получение списка контактов пользователя.
        """
        message = {
            ACTION: GET_CONTACTS,
            TIME: time(),
            FROM: self.user_name
        }
        client_log.debug(f'Запрос списка контактов от {self.user_name}')
        with sock_lock:
            send_message(self.socket, message)
            answer = get_message(self.socket)
        client_log.debug(f'Получен ответ на запрос списка контактов {answer}')
        if RESPONSE in answer and answer[RESPONSE] == 202:
            for contact in answer[ALERT]:
                self.db.add_contact(contact)
        else:
            client_log.error('Не удалось обновить список контактов.')

    @Log()
    def get_user_list(self):
        """
        Функция отправляет запрос серверу на получение списка известных пользователей.
        """
        client_log.debug(f'Запрос списка известных пользователей от {self.user_name}.')
        message = {
            ACTION: GET_USERS,
            TIME: time(),
            FROM: self.user_name
        }
        with sock_lock:
            send_message(self.socket, message)
            answer = get_message(self.socket)
        if RESPONSE in answer and answer[RESPONSE] == 202:
            self.db.add_all_users(answer[ALERT])
        else:
            client_log.error('Не удалось обновить список известных пользователей.')

    def key_request(self, user):
        """
        Запрашивает публичный ключ пользователя.
        """
        client_log.debug(f'Запрос публичного ключа для {user}')
        req = {
            ACTION: PUBLIC_KEY_REQUEST,
            TIME: time(),
            ACCOUNT_NAME: user
        }
        with sock_lock:
            send_message(self.socket, req)
            ans = get_message(self.socket)
        if RESPONSE in ans and ans[RESPONSE] == 511:
            return ans[DATA]
        else:
            client_log.error(f'Не удалось получить ключ пользователя {user}.')

    @Log()
    def add_contact(self, nickname):
        """
        Функция добавляет контакт в список контактов.
        """
        message = {
            ACTION: ADD_CONTACT,
            FROM: self.user_name,
            TIME: time(),
            LOGIN: nickname
        }
        client_log.debug(f'Отправлен запрос на добавление контакта {nickname} в список контактов.')

        with sock_lock:
            send_message(self.socket, message)
            self.read_response(get_message(self.socket))

    @Log()
    def del_contact(self, nickname):
        """
        Функция удаляет контакт из списка контактов.
        """
        message = {
            ACTION: DEL_CONTACT,
            FROM: self.user_name,
            TIME: time(),
            LOGIN: nickname
        }
        client_log.debug(f'Отправлен запрос на удаление контакта {nickname} из списка контактов.')

        with sock_lock:
            send_message(self.socket, message)
            self.read_response(get_message(self.socket))

    @Log()
    def create_user_message(self, recipient, message_text):
        """
        Функция формирует сообщение пользователя и отправляет его.
        """
        message = {
            ACTION: MSG,
            TIME: time(),
            FROM: self.user_name,
            TO: recipient,
            TEXT: message_text
        }
        client_log.debug(f'Создано сообщение от {self.user_name} для {recipient}')

        with sock_lock:
            send_message(self.socket, message)
            self.read_response(get_message(self.socket))
            client_log.info(f'Отрправлено сообщение {message}')

    @Log()
    def connection_shutdown(self):
        """
        Функция закрывает соединение с сервером.
        """
        self.running = False
        message = {
            ACTION: EXIT,
            TIME: time(),
            FROM: self.user_name
        }
        with sock_lock:
            try:
                send_message(self.socket, message)
            except OSError:
                pass
        client_log.debug('Завершение работы.')
        sleep(0.5)

    @Log()
    def read_response(self, message):
        """
        Функция принимает сообщение сервера, разбирает его
        и выводит соответствующий результат.
        """
        client_log.debug(f'Разбор ответа сервера: {message}')
        # Если пришол ответ сервера на запрос:
        if RESPONSE in message:
            if message[RESPONSE] == 200:
                client_log.info('Получен код подтверждения сервера 200')
                return
            elif message[RESPONSE] == 400:
                raise ServerError(f'{message[ERROR]}')
            else:
                client_log.debug(f'Принят неизвестный код подтверждения {message[RESPONSE]}')

        # Если пришло текстовое сообщение от пользователя:
        elif (ACTION in message and message[ACTION] == MSG
                and TIME in message and FROM in message
                and TEXT in message
                and TO in message and message[TO] == self.user_name):
            client_log.info(f'Получено сообщение {message}')
            self.new_message.emit(message)

    def run(self):
        """
        Основная функция-процесс для приёмки сообщений.
        """
        client_log.info(f'Запуск процесса для приёмки сообщений.')
        while self.running:
            sleep(1)
            with sock_lock:
                try:
                    self.socket.settimeout(0.5)
                    message = get_message(self.socket)
                except OSError as err:
                    if err.errno:
                        client_log.critical(f'Потеряно соединение с сервером.')
                        self.running = False
                        self.connection_lost.emit()
                except (ConnectionError, ConnectionAbortedError,
                        ConnectionResetError, json.JSONDecodeError, TypeError):
                    client_log.debug(f'Потеряно соединение с сервером.')
                    self.running = False
                    self.connection_lost.emit()
                else:
                    client_log.debug(f'Принято сообщение с сервера: {message}')
                    self.read_response(message)
                finally:
                    self.socket.settimeout(5)



"""
Графический интерфейс пользователя.
Содержит класс основного окна клиентского интерфейса.
Отображает:
* спискок контактов;
* историю сообщений с выбраным пользователем;
* поле ввода и кнопки отпраки сообщения.
Для создания графического интерфейса используется библиотека PyQt.
"""
import base64
import json
import logging
import sys

from Cryptodome.Cipher import PKCS1_OAEP
from Cryptodome.PublicKey import RSA
from PyQt5.QtCore import pyqtSlot, Qt
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QMainWindow, qApp, QMessageBox

from client.add_contact import AddContactDialog
from client.del_contact import DelContactDialog
from client.main_window_ui import Ui_MainClientWindow

sys.path.append('../')
from common.errors import ServerError
from common.variables import *

client_log = logging.getLogger('client')


class ClientMainWindow(QMainWindow):
    """
    Класс основного клиентского окна.
    """
    def __init__(self, database, connection, keys):
        super().__init__()
        self.database = database
        self.connection = connection

        self.decrypter = PKCS1_OAEP.new(keys)

        self.ui = Ui_MainClientWindow()
        self.ui.setupUi(self)

        self.ui.menu_exit.triggered.connect(qApp.exit)
        self.ui.btn_send.clicked.connect(self.send_message)
        # добавить контакт
        self.ui.btn_add_contact.clicked.connect(self.add_contact_window)
        self.ui.menu_add_contact.triggered.connect(self.add_contact_window)
        # Удалить контакт
        self.ui.btn_remove_contact.clicked.connect(self.delete_contact_window)
        self.ui.menu_del_contact.triggered.connect(self.delete_contact_window)

        # Дополнительные требующиеся атрибуты
        self.contacts_model = None
        self.history_model = None
        self.messages = QMessageBox()
        self.current_chat = None
        self.current_chat_key = None
        self.encryptor = None
        self.ui.list_messages.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.ui.list_messages.setWordWrap(True)

        # Даблклик по листу контактов отправляется в обработчик
        self.ui.list_contacts.doubleClicked.connect(self.select_active_user)

        self.contact_list_update()
        self.disable_input()
        self.show()

    def disable_input(self):
        """
        Функция делает поля ввода неактивными.
        """
        self.ui.label_new_message.setText('Для выбора получателя '
                                          'дважды кликните на нем в списке контактов.')
        self.ui.text_message.clear()
        if self.history_model:
            self.history_model.clear()

        self.ui.btn_clear.setDisabled(True)
        self.ui.btn_send.setDisabled(True)
        self.ui.text_message.setDisabled(True)

        self.encryptor = None
        self.current_chat = None
        self.current_chat_key = None

    def update_history(self):
        """
        Функция заполняет историю сообщений с выбранным пользователем.
        """
        # Получаем историю
        history = sorted(self.database.get_message_history(self.current_chat),
                         key=lambda msg: msg[2])

        # Создаём модель
        if not self.history_model:
            self.history_model = QStandardItemModel()
            self.ui.list_messages.setModel(self.history_model)
        self.history_model.clear()

        # Берём не более 20 последних записей.
        length = len(history)
        start_index = 0
        if length > 20:
            start_index = length - 20
        for i in range(start_index, length):
            msg = history[i]
            mess = QStandardItem(f'{msg[2].replace(microsecond=0)}\n'
                                 f'От {msg[0]} для {msg[1]}:\n'
                                 f'{msg[3]}')
            mess.setEditable(False)
            mess.setTextAlignment(Qt.AlignLeft)
            self.history_model.appendRow(mess)

        self.ui.list_messages.scrollToBottom()

    def select_active_user(self):
        """
        Функция обработчик двойного клика по контакту.
        """
        self.current_chat = self.ui.list_contacts.currentIndex().data()
        self.set_active_user()

    def set_active_user(self):
        """
        Функция устанавливает собеседника,
        получает его историю и активирует поле ввода сообщений.
        """
        try:
            self.current_chat_key = self.connection.key_request(self.current_chat)
            client_log.debug(f'Загружен открытый ключ для {self.current_chat}')
            if self.current_chat_key:
                self.encryptor = PKCS1_OAEP.new(RSA.import_key(self.current_chat_key))
        except (OSError, json.JSONDecodeError):
            self.current_chat_key = None
            self.encryptor = None
            client_log.debug(f'Не удалось получить ключ для {self.current_chat}')

        if not self.current_chat_key:
            self.messages.warning(
                self, 'Ошибка', 'Для выбранного пользователя нет ключа шифрования.')
            return

        self.ui.label_new_message.setText(f'Введите сообщенние для {self.current_chat}:')
        self.ui.btn_clear.setDisabled(False)
        self.ui.btn_send.setDisabled(False)
        self.ui.text_message.setDisabled(False)
        self.update_history()

    def contact_list_update(self):
        """
        Функция обновляет контакт лист.
        """
        contacts_list = self.database.get_contacts()
        self.contacts_model = QStandardItemModel()
        for i in contacts_list:
            item = QStandardItem(i)
            item.setEditable(False)
            self.contacts_model.appendRow(item)
        self.ui.list_contacts.setModel(self.contacts_model)

    def add_contact_window(self):
        """
        Функция добавления контакта.
        """
        global select_dialog
        select_dialog = AddContactDialog(self.connection, self.database)
        select_dialog.btn_ok.clicked.connect(lambda: self.add_contact_action(select_dialog))
        select_dialog.show()

    def add_contact_action(self, item):
        """
        Функция - обработчик добавления контактов.
        """
        contact = item.selector.currentText()
        self.add_contact(contact)
        item.close()

    def add_contact(self, name):
        """
        Функция добавляюет контакт в базу.
        """
        try:
            self.connection.add_contact(name)
        except ServerError as err:
            self.messages.critical(self, 'Ошибка сервера', err.text)
        except OSError as err:
            if err.errno:
                self.messages.critical(self, 'Ошибка', 'Потеряно соединение с сервером!')
                self.close()
            self.messages.critical(self, 'Ошибка', 'Таймаут соединения!')
        else:
            self.database.add_contact(name)
            new_contact = QStandardItem(name)
            new_contact.setEditable(False)
            self.contacts_model.appendRow(new_contact)
            client_log.info(f'Успешно добавлен контакт {new_contact}')
            self.messages.information(self, 'Успех', 'Контакт успешно добавлен.')

    def delete_contact_window(self):
        """
        Создать окно удаления контакта.
        """
        global remove_dialog
        remove_dialog = DelContactDialog(self.database)
        remove_dialog.btn_ok.clicked.connect(lambda: self.delete_contact(remove_dialog))
        remove_dialog.show()

    def delete_contact(self, item):
        """
        Функция удаления контакта
        """
        selected = item.selector.currentText()
        try:
            self.connection.del_contact(selected)
        except ServerError as err:
            self.messages.critical(self, 'Ошибка сервера', err.text)
        except OSError as err:
            if err.errno:
                self.messages.critical(self, 'Ошибка', 'Потеряно соединение с сервером!')
                self.close()
            self.messages.critical(self, 'Ошибка', 'Таймаут соединения!')
        else:
            self.database.del_contact(selected)
            self.contact_list_update()
            client_log.info(f'Успешно удалён контакт {selected}')
            self.messages.information(self, 'Успех', 'Контакт успешно удалён.')
            item.close()

            if selected == self.current_chat:
                self.current_chat = None
                self.disable_input()

    def send_message(self):
        """
        Функция отправки собщения пользователю.
        """
        message_text = self.ui.text_message.toPlainText()
        self.ui.text_message.clear()
        if not message_text:
            return

        # Штфруем сообщение
        message_text_encrypted = self.encryptor.encrypt(message_text.encode('utf8'))
        message_text_enc_base64 = base64.b64encode(message_text_encrypted)

        try:
            self.connection.create_user_message(self.current_chat,
                                                message_text_enc_base64.decode('ascii'))
        except ServerError as err:
            self.messages.critical(self, 'Ошибка', err.text)
        except OSError as err:
            if err.errno:
                self.messages.critical(self, 'Ошибка', 'Потеряно соединение с сервером!')
                self.close()
            self.messages.critical(self, 'Ошибка', 'Таймаут соединения!')
        except (ConnectionResetError, ConnectionAbortedError):
            self.messages.critical(self, 'Ошибка', 'Потеряно соединение с сервером!')
            self.close()
        else:
            self.database.save_message(self.connection.user_name,
                                       self.current_chat,
                                       message_text)
            client_log.debug(f'Отправлено сообщение для {self.current_chat}: {message_text}')
            self.update_history()

    @pyqtSlot(dict)
    def message(self, message):
        """
        Слот приёма нового сообщения
        """
        encrypted_message = base64.b64decode(message[TEXT])
        try:
            decrypted_message = self.decrypter.decrypt(encrypted_message)
        except (ValueError, TypeError):
            self.messages.warning(self, 'Ошибка', 'Не удалось декодировать сообщение.')
            return

        self.database.save_message(message[FROM],
                                   message[TO],
                                   decrypted_message.decode('utf8'))

        sender = message[FROM]
        if sender == self.current_chat:
            self.update_history()
        else:
            if self.database.check_contact(sender):
                # Если пользователь в контактах
                if self.messages.question(self, 'Новое сообщение',
                                          f'Получено новое сообщение от {sender}, открыть чат с ним?',
                                          QMessageBox.Yes,
                                          QMessageBox.No) == QMessageBox.Yes:
                    self.current_chat = sender
                    self.set_active_user()
            else:
                # Если такого пользователя нет
                if self.messages.question(self, 'Новое сообщение',
                                          f'Получено новое сообщение от {sender}.\n '
                                          f'Данного пользователя нет в вашем контакт-листе.\n'
                                          f' Добавить в контакты и открыть чат с ним?',
                                          QMessageBox.Yes,
                                          QMessageBox.No) == QMessageBox.Yes:
                    self.add_contact(sender)
                    self.current_chat = sender
                    self.database.save_message(message[FROM],
                                               message[TO],
                                               decrypted_message.decode('utf8'))
                    self.set_active_user()

    @pyqtSlot()
    def connection_lost(self):
        """
        Слот потери соединения.
        """
        self.messages.warning(self, 'Сбой соединения', 'Потеряно соединение с сервером.')
        self.close()

    def make_connection(self, trans_obj):
        trans_obj.new_message.connect(self.message)
        trans_obj.connection_lost.connect(self.connection_lost)
